# ==================================================================
#       文 件 名: admin.py
#       概    要: 后台管理基类
#       作    者: IT小强 
#       创建时间: 6/15/20 4:21 PM
#       修改时间: 
#       copyright (c) 2016 - 2020 mail@xqitw.cn
# ==================================================================

from django.contrib import admin

from mptt.admin import MPTTModelAdmin


class ModelAdmin(admin.ModelAdmin):
    """
    后台管理基类
    """


class ModelAdminWithUser(ModelAdmin):
    """
    后台管理基类（自动写入创建用户ID和更新用户ID）
    """

    def save_model(self, request, obj, form, change):
        """
        Given a model instance save it to the database.
        自动写入创建用户ID和更新用户ID
        :param request:
        :param obj:
        :param form:
        :param change:
        :return:
        """
        if request.user:
            user = request.user
            obj.updated_user = user
            if not change:
                obj.created_user = user
        super().save_model(request, obj, form, change)


class ModelAdminMPTT(MPTTModelAdmin):
    """
    后台管理基类(支持树形结构)
    """


class ModelAdminMPTTWithUser(ModelAdminMPTT):
    """
     后台管理基类(支持树形结构)（自动写入创建用户ID和更新用户ID）
    """

    def save_model(self, request, obj, form, change):
        """
        Given a model instance save it to the database.
        自动写入创建用户ID和更新用户ID
        :param request:
        :param obj:
        :param form:
        :param change:
        :return:
        """
        if request.user:
            user = request.user
            obj.updated_user = user
            if not change:
                obj.created_user = user
        super().save_model(request, obj, form, change)



# ==================================================================
#       文 件 名: model.py
#       概    要: 模型基类
#       作    者: IT小强 
#       创建时间: 6/8/20 2:38 PM
#       修改时间: 
#       copyright (c) 2016 - 2020 mail@xqitw.cn
# ==================================================================

from json import dumps

from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from mptt.models import MPTTModel as BaseMPTTModel

from .helper import get_kelove_databases_settings, STATUS_CHOICES


class Model(models.Model):
    """
    数据模型基类
    """

    class Meta:
        abstract = True


class CreatedUserModel(Model):
    """
    添加创建用户外键
    """

    # 创建用户
    created_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('created_user'),
        related_name="%(app_label)s_%(class)s_created_user_set",
        on_delete=get_kelove_databases_settings('FOREIGN_DELETE_TYPE', models.PROTECT),
        db_constraint=get_kelove_databases_settings('DB_CONSTRAINT_USER', False),
        null=True,
        blank=True,
        editable=get_kelove_databases_settings('USER_EDITABLE', False),
        default=None
    )

    class Meta:
        abstract = True


class UpdatedUserModel(Model):
    """
    添加更新用户外键
    """

    # 更新用户
    updated_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('updated_user'),
        related_name="%(app_label)s_%(class)s_updated_user_set",
        on_delete=get_kelove_databases_settings('FOREIGN_DELETE_TYPE', models.PROTECT),
        db_constraint=get_kelove_databases_settings('DB_CONSTRAINT_USER', False),
        null=True,
        blank=True,
        editable=get_kelove_databases_settings('USER_EDITABLE', False),
        default=None
    )

    class Meta:
        abstract = True


class CreatedTimeModel(Model):
    """
    添加创建时间
    """

    # 创建时间
    created_time = models.DateTimeField(verbose_name=_('created_time'), auto_now_add=True, editable=True)

    class Meta:
        abstract = True


class UpdatedTimeModel(Model):
    """
    添加更新时间
    """

    # 更新时间
    updated_time = models.DateTimeField(verbose_name=_('updated_time'), auto_now=True)

    class Meta:
        abstract = True


class StatusModel(Model):
    """
    添加状态字段
    """

    STATUS_CHOICES = get_kelove_databases_settings('STATUS_CHOICES', STATUS_CHOICES)

    # 状态
    status = models.IntegerField(
        verbose_name=_('status'),
        null=False,
        default=0,
        choices=STATUS_CHOICES,
        help_text=_('status') + dumps(dict(STATUS_CHOICES), ensure_ascii=False)
    )

    class Meta:
        abstract = True


class EnabledModel(Model):
    """
    添加是否启用字段
    """

    # 是否启用
    enabled = models.BooleanField(
        verbose_name=_('enabled'),
        default=False,
        db_index=True,
    )

    class Meta:
        abstract = True


class SortModel(Model):
    """
    添加排序字段
    """

    # 排序
    sort = models.IntegerField(
        verbose_name=_('sort'),
        null=False,
        default=0,
    )

    class Meta:
        abstract = True


class NoUserModel(CreatedTimeModel, UpdatedTimeModel, SortModel, StatusModel, EnabledModel):
    """
    数据模型基类（不包括创建用户、更新用户字段）
    """

    class Meta:
        abstract = True


class NoTimeModel(CreatedUserModel, UpdatedUserModel, SortModel, StatusModel, EnabledModel):
    """
    数据模型基类（不包括创建时间、更新时间字段）
    """

    class Meta:
        abstract = True


class OnlyTimeModel(CreatedTimeModel, UpdatedTimeModel):
    """
    数据模型基类（仅包含时间字段）
    """

    class Meta:
        abstract = True


class OnlyUserModel(CreatedUserModel, UpdatedUserModel):
    """
    数据模型基类（仅包含创建、更新用户字段）
    """

    class Meta:
        abstract = True


class AllModel(OnlyUserModel, OnlyTimeModel, StatusModel, EnabledModel, SortModel):
    """
    数据模型基类(添加全部公用字段)
    """

    class Meta:
        abstract = True


class MPTTModel(BaseMPTTModel):
    """
    树形数据模型基类
    """

    level = models.PositiveIntegerField(verbose_name=_('level'), editable=False)

    class Meta:
        abstract = True



from abc import ABC

from django.db.backends.mysql.schema import DatabaseSchemaEditor as MySqlDatabaseSchemaEditor


class DatabaseSchemaEditor(MySqlDatabaseSchemaEditor, ABC):

    def column_sql(self, model, field, include_default=False):
        """
        重写 字段sql生成方法
        :param model:
        :param field:
        :param include_default:
        :return:
        """

        # 字段默认值是否写入到sql语句中处理，可在settings.py中配置
        include_default_fun = self.connection.settings_dict.get('INCLUDE_DEFAULT', None)

        if isinstance(include_default_fun, bool):
            include_default = include_default_fun
        elif include_default_fun is not None:
            include_default = include_default_fun(model, field, include_default, self.connection)

        # 生成sql
        sql, params = super().column_sql(model, field, include_default)

        if not isinstance(sql, str):
            return sql, params

        # 写入字段注释
        if field.help_text:
            sql += " COMMENT '%s'" % field.help_text
        elif field.verbose_name:
            sql += " COMMENT '%s'" % field.verbose_name
        return sql, params

    def table_sql(self, model):
        """
        重写表sql生成方法
        :param model:
        :return:
        """
        sql, params = super().table_sql(model)
        if model._meta.verbose_name:
            sql += " COMMENT '%s'" % model._meta.verbose_name
        return sql, params



# ==================================================================
#       文 件 名: doc.py
#       概    要: 生成数据库文档
#       作    者: IT小强 
#       创建时间: 6/9/20 5:52 PM
#       修改时间: 
#       copyright (c) 2016 - 2020 mail@xqitw.cn
# ==================================================================

from ..base import doc


class Doc(doc.Doc):
    """
    生成数据库文档
    """

    COLUMNS_SQL = 'SHOW FULL COLUMNS FROM `{table_name}`'

    def get_db_fields_info(self, table_name, cur_fields):
        """
        获取数据库中的字段信息
        :param table_name:
        :param cur_fields:
        :return:
        """

        sql = self.COLUMNS_SQL.format(table_name=table_name)
        self.cursor.execute(sql)
        for filed_info in self.cursor.fetchall():
            filed_name = filed_info[0]
            # db_default_value = filed_info[5]
            # default_value = cur_fields[filed_name]['default']
            # comment = cur_fields[filed_name]['comment']

            cur_filed_info = {
                'field': filed_name,
                'type': filed_info[1],
                'collation': filed_info[2],
                'null': filed_info[3],
                'key': filed_info[4],
                # 'default': db_default_value if db_default_value else default_value,
                'extra': filed_info[6],
                'privileges': filed_info[7],
                # 'comment': comment if comment else filed_info[8]
            }
            cur_fields[filed_name].update(cur_filed_info)
        return cur_fields



# ==================================================================
#       文 件 名: helper.py
#       概    要: 助手工具
#       作    者: IT小强 
#       创建时间: 6/8/20 2:37 PM
#       修改时间: 
#       copyright (c) 2016 - 2020 mail@xqitw.cn
# ==================================================================

from importlib import import_module

from django.conf import settings

STATUS_CHOICES: list = [(-1, '草稿'), (0, '待审'), (1, '通过'), (2, '驳回')]


def load_object(path: str):
    """
    Load an object given its absolute object path, and return it.
    object can be a class, function, variable or an instance.
    path ie: 'scrapy.downloadermiddlewares.redirect.RedirectMiddleware'
    """

    dot = path.rindex('.')
    module, name = path[:dot], path[dot + 1:]
    mod = import_module(module)
    return getattr(mod, name)


def get_kelove_databases_settings(key: str = None, default=None):
    """
    获取配置
    :param key:
    :param default:
    :return:
    """
    try:
        kelove_databases_settings = settings.KELOVE_DATABASES
    except AttributeError:
        kelove_databases_settings = {}

    if not isinstance(kelove_databases_settings, dict):
        kelove_databases_settings = {}

    default_kelove_databases_settings = {
        'FOREIGN_DELETE_TYPE': 'django.db.models.deletion.PROTECT',
        'DB_CONSTRAINT': False,
        'DB_CONSTRAINT_USER': False,
        'USER_EDITABLE': False,
        'STATUS_CHOICES': STATUS_CHOICES
    }
    kelove_databases_settings = {**default_kelove_databases_settings, **kelove_databases_settings}

    if isinstance(kelove_databases_settings['FOREIGN_DELETE_TYPE'], str):
        kelove_databases_settings['FOREIGN_DELETE_TYPE'] = load_object(kelove_databases_settings['FOREIGN_DELETE_TYPE'])

    if key is None:
        return kelove_databases_settings
    else:
        return kelove_databases_settings.get(key, default)



# ==================================================================
#       文 件 名: doc.py
#       概    要: 生成数据库文档
#       作    者: IT小强 
#       创建时间: 6/9/20 5:52 PM
#       修改时间: 
#       copyright (c) 2016 - 2020 mail@xqitw.cn
# ==================================================================

from django.db import connection
from django.apps import apps


class Doc:
    """
    生成数据库文档
    """

    def __init__(self):
        self.cursor = connection.cursor()

    def get_db_fields_info(self, table_name, cur_fields):
        """
        获取数据库中的字段信息
        :param table_name:
        :param cur_fields:
        :return:
        """
        pass

    def get_fields(self, table_name, fields):
        """
        获取数据表全部字段信息
        :param fields:
        :param table_name:
        :return:
        """
        cur_fields = {}

        # 获取模型信息
        for item in fields:
            cur_fields[item.column] = {
                # 字段注释
                'comment': item.help_text if item.help_text else item.verbose_name,
                # 字段默认值
                'default': item.get_default(),
                # 获取关联表
                'relation_table_name': item.related_model._meta.db_table if item.is_relation else ''
            }

        # 获取数据库查询信息并合并
        return self.get_db_fields_info(table_name, cur_fields)

    def get_tables(self, app):
        """
        获取指定模块下的全部模型
        :param app:
        :return:
        """
        data = {}
        for model in apps.get_app_config(app).get_models():
            table_name = model._meta.db_table  # 表名

            cur_table_info = {
                'name': table_name,
                'title': model._meta.verbose_name,
                'fields': self.get_fields(table_name, model._meta.fields)
            }
            if cur_table_info['fields']:
                data[cur_table_info['name']] = cur_table_info
        return data

    def get_database_doc_data(self, apps_list=None):
        """
        获取应用数据库文档数据
        :param apps_list:
        :return:
        """
        data = {}

        if apps_list is None:
            apps_list = [item for item in apps.all_models.keys()]

        for app in apps_list:
            cur_app_info = {
                'name': app,
                'title': apps.get_app_config(app).verbose_name,
                'tables': self.get_tables(app)
            }
            if cur_app_info['tables']:
                data[app] = cur_app_info
        return data



"""
MySQL database backend for Django.
"""

from django.db.backends.mysql.base import *

from .schema import DatabaseSchemaEditor as MySqlDatabaseSchemaEditor


class DatabaseWrapper(DatabaseWrapper):
    SchemaEditorClass = MySqlDatabaseSchemaEditor



# ==================================================================
#       文 件 名: setup.py
#       概    要: 扩展包构建
#       作    者: IT小强 
#       创建时间: 5/28/20 9:57 PM
#       修改时间: 
#       copyright (c) 2016 - 2020 mail@xqitw.cn
# ==================================================================

from json import loads

from os import path as path

from setuptools import setup, find_packages

from django_kelove_db import PACKAGE_VERSION

base_directory = path.abspath(path.dirname(__file__))

# 获取全部的包信息
all_packages = find_packages()

# 移除 django_kelove 包
try:
    all_packages.remove('django_kelove')
except ValueError:
    pass


def read_file(filename):
    """
    读取文件内容
    :param filename: 文件名
    :return:
    """
    with open(path.join(base_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
        return long_description


def read_requirements(filename='Pipfile.lock'):
    """
    获取依赖
    :param filename: 文件名
    :return:
    """
    requires = loads(read_file(filename))
    requires = requires.get('default', {})
    requires = [(k + v.get('version', '')) for k, v in requires.items()]
    return requires


setup(

    name='django-kelove-db',

    python_requires='>=3.7.0',

    version=PACKAGE_VERSION,

    description="Django数据库迁移优化（目前只完善MySQL）",

    long_description=read_file('README.md'),

    long_description_content_type="text/markdown",

    author="IT小强xqitw.cn",

    author_email='mail@xqitw.cn',

    url='https://e.coding.net/xqitw/django-kelove/django-kelove-db.git',

    packages=all_packages,

    install_requires=read_requirements(),

    include_package_data=True,

    license="Apache-2.0",

    keywords=['django', 'mysql', 'comment'],

    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: Chinese (Simplified)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



# ==================================================================
#       文 件 名: views.py
#       概    要: 视图
#       作    者: IT小强 
#       创建时间: 6/9/20 5:58 PM
#       修改时间: 
#       copyright (c) 2016 - 2020 mail@xqitw.cn
# ==================================================================

from django.shortcuts import render
from django.views.generic import View

from .backends.mysql.doc import Doc
from .util import helper


class MySQLDatabaseDoc(View):
    """
    在线数据库文档
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.doc_title = helper.get_kelove_databases_settings('DOC_TITLE', '数据库设计文档')

    def get(self, request):
        """
        处理GET 请求
        :param request:
        :return:
        """
        doc = Doc()
        doc_data = doc.get_database_doc_data()

        return render(request, 'django_kelove_db/doc_db.html', {"apps": doc_data, 'title': self.doc_title})
